import 'package:valid_value_objects/src/value_exceptions.dart';

enum Boundary { before, after, inRange }

/// Class that stores a lower and/or an upper [DateTime] instance
/// to check if the given dates are in range.
class BoundaryDateTime {
  final DateTime? lowerBound;
  final DateTime? upperBound;

  /// Either [lowerBound] or [upperBound] can be null, but not both of them.
  ///
  /// Omitting one of the value represents no boundary.
  factory BoundaryDateTime({DateTime? lowerBound, DateTime? upperBound}) {
    if (lowerBound == null && upperBound == null) {
      throw const RequiredValueException();
    }
    if (lowerBound != null &&
        upperBound != null &&
        lowerBound.isAfter(upperBound)) {
      throw InvalidValueException(
        [lowerBound, upperBound],
        code: 'inalid-range',
        message: 'Lowerbound DateTime parameter must be before upperbound.',
      );
    }

    return BoundaryDateTime._(lowerBound, upperBound);
  }

  BoundaryDateTime._(this.lowerBound, this.upperBound);

  /// If [dateTime] is
  /// - before [lowerBound] then [Boundary.before] is returned
  /// - after [upperBound] then [Boundary.after] is returned
  /// - between [lowerBound] and [upperBound] then [Boundary.after] is returned
  Boundary boundary(DateTime dateTime) {
    if (lowerBound != null && dateTime.isBefore(lowerBound!)) {
      return Boundary.before;
    }
    if (upperBound != null && dateTime.isAfter(upperBound!)) {
      return Boundary.after;
    }

    return Boundary.inRange;
  }

  /// Checks if [dateTime] is in between [lowerBound] and [upperBound]
  bool isInRange(DateTime dateTime) {
    return boundary(dateTime) == Boundary.inRange;
  }

  @override
  bool operator ==(Object o) {
    if (identical(this, o)) {
      return true;
    }

    return o is BoundaryDateTime &&
        o.lowerBound == lowerBound &&
        o.upperBound == o.upperBound;
  }

  @override
  int get hashCode => lowerBound.hashCode ^ upperBound.hashCode;

  /// Returns: '${lowerBound},${upperBound}'
  @override
  String toString() {
    return '${lowerBound},${upperBound}';
  }
}
