import '../src/helper.dart';
import '../src/value_exceptions.dart';
import '../src/value_object.dart';

class EmailAddress extends ValueObject<String> {
  /// Default value: `email`
  static String key = 'email';

  // Validation str taken from here:
  // https://github.com/dart-league/validators/blob/master/lib/validators.dart
  static const defaultRegex =
      r"^((([a-z]|\d|[!#\$%&'\*\+\-\/=\?\^_`{\|}~]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])+(\.([a-z]|\d|[!#\$%&'\*\+\-\/=\?\^_`{\|}~]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])+)*)|((\x22)((((\x20|\x09)*(\x0d\x0a))?(\x20|\x09)+)?(([\x01-\x08\x0b\x0c\x0e-\x1f\x7f]|\x21|[\x23-\x5b]|[\x5d-\x7e]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(\\([\x01-\x09\x0b\x0c\x0d-\x7f]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]))))*(((\x20|\x09)*(\x0d\x0a))?(\x20|\x09)+)?(\x22)))@((([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])([a-z]|\d|-|\.|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])*([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])))\.)+(([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])([a-z]|\d|-|\.|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])*([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])))$";

  static bool _defaultValidator(String str) =>
      RegExp(defaultRegex).hasMatch(str.toLowerCase());

  static ValidCallback? _customValidator;

  /// Define your own validator if you're not satisfied with the default one.
  ///
  /// Object creations and [validate] function validations
  /// will be validated by your custom [validation]
  static void customValidator(ValidCallback? validation) =>
      _customValidator = validation;

  /// Checks if [str] is an email.
  ///
  /// Validation happens by the default or by your custom
  /// (if defined) validator function.
  static bool validate(String str) =>
      _customValidator?.call(str) ?? _defaultValidator(str);

  /// One of the following [ValueException] maybe thrown:
  /// - RequiredValueException
  /// - InvalidValueException
  factory EmailAddress(String str) {
    if (str.isEmpty) {
      throw const RequiredValueException();
    } else if (validate(str)) {
      return EmailAddress._(str);
    }

    throw InvalidValueException(str, message: 'Invalid email address.');
  }

  /// Returns the string **before** the @ character.
  ///
  /// For example: **xy@gmail.com** returns only **xy**
  String get local => toString().split('@')[0];

  /// Returns the string **after** the @ character.
  ///
  /// For example: **xy@gmail.com** returns only **gmail.com**
  String get domain => toString().split('@')[1];

  const EmailAddress._(String value) : super(value);

  /// [map] can be a `Map`, `String` or `ValueObject`
  ///
  /// One of the following [ValueException] maybe thrown:
  /// - RequiredValueException
  /// - InvalidValueException
  factory EmailAddress.fromJson(dynamic map) {
    if (map is Map) {
      final flatMap = flattenMap(map);
      return EmailAddress(flatMap[key]);
    }
    return EmailAddress(map.toString());
  }

  Map<String, String> toJson() => {key: toString()};
}
