import '../src/helper.dart';
import '../valid_value_objects.dart';

class FirstName extends ValueObject<String> {
  /// Default value: `fistName`
  static String key = 'firstName';

  static ValidCallback? _customValidator;

  /// Define your own validator if you're not satisfied with the default one.
  ///
  /// Object creations and [validate] function validations
  /// will be validated by your custom [validation]
  static void customValidator(ValidCallback? validation) =>
      _customValidator = validation;

  /// Checks if [str] is a first name.
  ///
  /// Validation happens by the default or by your custom
  /// (if defined) validator function.
  static bool validate(String str) =>
      _customValidator?.call(str) ?? _defaultValidator(str);

  /// One of the following [ValueException] maybe thrown:
  /// - RequiredValueException
  /// - InvalidValueException
  factory FirstName(String value) {
    if (value.isEmpty) {
      throw const RequiredValueException();
    } else if (!validate(value)) {
      throw InvalidValueException(value);
    }
    return FirstName._(value);
  }

  const FirstName._(String value) : super(value);

  /// One of the following [ValueException] maybe thrown:
  /// - RequiredValueException
  /// - InvalidValueException
  factory FirstName.fromJson(dynamic map) {
    if (map is Map) {
      final flatMap = flattenMap(map);
      return FirstName(flatMap[key]);
    }
    return FirstName(map.toString());
  }

  Map<String, String> toJson() => {key: toString()};
}

class LastName extends ValueObject<String> {
  /// Default value: `lastName`
  static String key = 'lastName';

  static ValidCallback? _customValidator;

  /// Define your own validator if you're not satisfied with the default one.
  ///
  /// Object creations and [validate] function validations
  /// will be validated by your custom [validation]
  static void customValidator(ValidCallback? validation) =>
      _customValidator = validation;

  /// Checks if [str] is a last name.
  ///
  /// Validation happens by the default or by your custom
  /// (if defined) validator function.
  static bool validate(String str) =>
      _customValidator?.call(str) ?? _defaultValidator(str);

  /// One of the following [ValueException] maybe thrown:
  /// - RequiredValueException
  /// - InvalidValueException
  factory LastName(String value) {
    if (value.isEmpty) {
      throw const RequiredValueException();
    } else if (!validate(value)) {
      throw InvalidValueException(value);
    }
    return LastName._(value);
  }

  const LastName._(String value) : super(value);

  /// One of the following [ValueException] maybe thrown:
  /// - RequiredValueException
  /// - InvalidValueException
  factory LastName.fromJson(dynamic map) {
    if (map is Map) {
      final flatMap = flattenMap(map);
      return LastName(flatMap[key]);
    }
    return LastName(map.toString());
  }

  Map<String, String> toJson() => {key: toString()};
}

class MiddleName extends ValueObject<String> {
  /// Default value: `middleName`
  static String key = 'middleName';

  static ValidCallback? _customValidator;

  /// Define your own validator if you're not satisfied with the default one.
  ///
  /// Object creations and [validate] function validations
  /// will be validated by your custom [validation]
  static void customValidator(ValidCallback? validation) =>
      _customValidator = validation;

  /// Checks if [str] is a middle name.
  ///
  /// Validation happens by the default or by your custom
  /// (if defined) validator function.
  static bool validate(String str) =>
      _customValidator?.call(str) ?? _defaultValidator(str);

  /// One of the following [ValueException] maybe thrown:
  /// - RequiredValueException
  /// - InvalidValueException
  factory MiddleName(String value) {
    if (value.isEmpty) {
      throw const RequiredValueException();
    } else if (!validate(value)) {
      throw InvalidValueException(value);
    }
    return MiddleName._(value);
  }

  const MiddleName._(String value) : super(value);

  /// One of the following [ValueException] maybe thrown:
  /// - RequiredValueException
  /// - InvalidValueException
  factory MiddleName.fromJson(dynamic map) {
    if (map is Map) {
      final flatMap = flattenMap(map);
      return MiddleName(flatMap[key]);
    }
    return MiddleName(map.toString());
  }

  Map<String, String> toJson() => {key: toString()};
}

class Name extends ValueObject {
  /// Default value: `name`
  static String key = 'name';

  static ValidCallback? _customValidator;

  /// Define your own validator if you're not satisfied with the default one.
  ///
  /// Object creations and [validate] function validations
  /// will be validated by your custom [validation]
  static void customValidator(ValidCallback? validation) =>
      _customValidator = validation;

  /// Checks if [str] is a name.
  ///
  /// Validation happens by the default or by your custom
  /// (if defined) validator function.
  static bool validate(String str) =>
      _customValidator?.call(str) ?? _defaultValidator(str);

  /// One of the following [ValueException] maybe thrown:
  /// - RequiredValueException
  /// - InvalidValueException
  factory Name(String value) {
    if (value.isEmpty) {
      throw const RequiredValueException();
    } else if (!validate(value)) {
      throw InvalidValueException(value);
    }
    return Name._(value);
  }

  const Name._(String value) : super(value);

  /// One of the following [ValueException] maybe thrown:
  /// - RequiredValueException
  /// - InvalidValueException
  factory Name.fromJson(dynamic map) {
    if (map is Map) {
      final flatMap = flattenMap(map);
      return Name(flatMap[key]);
    }
    return Name(map.toString());
  }

  Map<String, String> toJson() => {key: toString()};
}

bool _defaultValidator(String str) {
  return str.length > 1;
}
