import '../src/helper.dart';
import '../src/value_exceptions.dart';
import '../src/value_object.dart';

class Password extends ValueObject<String> {
  /// Default value: `password`
  static String key = 'password';

  static int minChar = 8;
  static int maxChar = 255;

  static bool mustContainNumeric = true;
  static bool mustContainLowerChar = true;
  static bool mustContainUpperChar = true;
  static bool mustContainSpecialChar = false;

  /// One of the following [ValueException] maybe thrown:
  /// - RequiredValueException
  /// - TooShortValueException
  /// - TooLongValueException
  /// - InvalidValueException
  factory Password(String value) {
    if (value.isEmpty) {
      throw const RequiredValueException();
    } else if (value.length < minChar) {
      throw TooShortValueException(value);
    } else if (value.length > maxChar) {
      throw TooLongValueException(value);
    } else if (Password.mustContainLowerChar && !hasLowerChar(value)) {
      throw InvalidValueException(
        value,
        code: 'must-contain-lower',
        message: 'Password must contain a lower case letter.',
      );
    } else if (Password.mustContainUpperChar && !hasUpperChar(value)) {
      throw InvalidValueException(
        value,
        code: 'must-contain-upper',
        message: 'Password must contain an upper case letter.',
      );
    } else if (Password.mustContainNumeric && !hasNumeric(value)) {
      throw InvalidValueException(
        value,
        code: 'must-contain-number',
        message: 'Password must contain a number.',
      );
    } else if (Password.mustContainSpecialChar && !hasSpecialChar(value)) {
      throw InvalidValueException(
        value,
        code: 'must-have-special',
        message: 'Password must contain a special character.',
      );
    }

    return Password._(value);
  }

  const Password._(String value) : super(value);

  /// [map] can be a `Map`, `String` or `ValueObject`
  ///
  /// One of the following [ValueException] maybe thrown:
  /// - RequiredValueException
  /// - InvalidValueException
  factory Password.fromJson(dynamic map) {
    if (map is Map) {
      final flatMap = flattenMap(map);
      return Password(flatMap[key]);
    }
    return Password(map.toString());
  }

  Map<String, String> toJson() => {key: toString()};

  static void setRequirements({
    int minChar = 6,
    int maxChar = 255,
    bool mustContainLowerChar = true,
    bool mustContainUpperChar = true,
    bool mustContainNumeric = true,
    bool mustContainSpecialChar = false,
  }) {
    Password.minChar = minChar;
    Password.maxChar = maxChar;
    Password.mustContainLowerChar = mustContainLowerChar;
    Password.mustContainUpperChar = mustContainUpperChar;
    Password.mustContainNumeric = mustContainUpperChar;
    Password.mustContainSpecialChar = mustContainSpecialChar;
  }
}
