import 'package:meta/meta.dart';
import '../src/helper.dart';
import '../src/value_exceptions.dart';
import '../src/value_object.dart';

@immutable
class PhoneNumber extends ValueObject<String> {
  /// Default value: `phone`
  static String key = 'phone';

  // Phone regex pattern's code taken from here: https://github.com/agungnursatria/regexpattern
  /// Phone Number regex
  /// Must started by either, "0", "+", "+XX <X between 2 to 4 digit>", "(+XX <X between 2 to 3 digit>)"
  /// Can add whitespace separating digit with "+" or "(+XX)"
  /// Example: 05555555555, +555 5555555555, (+123) 5555555555, (555) 5555555555, +5555 5555555555
  static const defaultRegex =
      r'^(0|\+|(\+[0-9]{2,4}|\(\+?[0-9]{2,4}\)) ?)([0-9]*|\d{2,4}-\d{2,4}(-\d{2,4})?)$';

  static bool _defaultValidator(String str) =>
      RegExp(defaultRegex).hasMatch(str.toLowerCase());

  static ValidCallback? _customValidator;

  /// Define your own validator if you're not satisfied with the default one.
  ///
  /// Object creations and [validate] function validations
  /// will be validated by your custom [validation]
  static void customValidator(ValidCallback? validation) =>
      _customValidator = validation;

  /// Checks if [str] is a phone number.
  ///
  /// Validation happens by the default or by your custom
  /// (if defined) validator function.
  static bool validate(String str) =>
      _customValidator?.call(str) ?? _defaultValidator(str);

  /// One of the following [ValueException] maybe thrown:
  /// - RequiredValueException
  /// - InvalidValueException
  factory PhoneNumber(String str) {
    if (str.isEmpty) {
      throw const RequiredValueException();
    } else if (validate(str)) {
      return PhoneNumber._(str);
    }
    throw InvalidValueException(str);
  }

  const PhoneNumber._(String value) : super(value);

  /// [map] can be a `Map`, `String` or `ValueObject`
  ///
  /// One of the following [ValueException] maybe thrown:
  /// - RequiredValueException
  /// - InvalidValueException
  factory PhoneNumber.fromJson(dynamic map) {
    if (map is Map) {
      final flatMap = flattenMap(map);
      return PhoneNumber(flatMap[key]);
    }
    return PhoneNumber(map.toString());
  }

  Map<String, String> toJson() => {key: toString()};
}
