import '../src/helper.dart';
import '../src/value_object.dart';

/// Generates a unique id:
///
/// *Example:* `2020-10-31 12:03:33.250111*b363c230e15c`
class UniqueId extends ValueObject<String> {
  /// Default value: `id`
  static String key = 'id';

  /// Creates a [UniqueId] by concatenating `DateTime.now()` with a random string separated by `*`
  ///
  /// Example: `2020-10-31 11:43:44.551947*LXoF9A`
  factory UniqueId() {
    final id = '${DateTime.now().toString()}*${getRandString(4)}';

    return UniqueId._(id);
  }

  const UniqueId._(String id) : super(id);

  /// Make sure to only use this constructor if [id] is unique or it is already created
  ///
  /// Otherwise always use `UniqueId()` constructor
  factory UniqueId.withUniqueId(String id) {
    assert(id.isNotEmpty);
    Uri.decodeComponent(id);
    return UniqueId._(id);
  }

  /// Returns the create time of this [UniqueId]
  ///
  /// Example: `DateTime(2020-10-31 11:43:44.551947)`
  ///
  /// **WARNING:** A null value maybe returned if the [UniqueId] instance is created by using the [withUniqueId] constructor
  DateTime? get createTime => DateTime.tryParse(super.toString().split('*')[0]);

  /// [map] can be a `Map`, `String` or `ValueObject`
  factory UniqueId.fromJson(dynamic map) {
    if (map is Map) {
      final flatMap = flattenMap(map);
      return UniqueId.withUniqueId(flatMap[key]);
    }
    return UniqueId.withUniqueId(map.toString());
  }

  Map<String, String> toJson() => {key: toString()};
}
